import re
import os
import yaml
import boto3
import logging
import requests
import posixpath
import unicodedata
from typing import Dict
from fmbt import globals
from transformers import AutoTokenizer
from botocore.exceptions import NoCredentialsError


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

_tokenizer = None

# initializing tokenizer in case it is encoded as none type 
def _initialize_tokenizer():
    global _tokenizer
    try:
        if _tokenizer is None:
            ## use normal tokenizer -- change this variable LOCAL_TOKENIZER_DIR
            local_dir = globals.TOKENIZER
            # Check if the tokenizer files exist locally
            if not os.path.exists(local_dir):
                # If not, download from S3
                bucket_name = globals.READ_BUCKET_NAME ## DO IT from config 
                prefix = globals.TOKENIZER_DIR_S3
                _download_from_s3(bucket_name, prefix, local_dir)
            # Load the tokenizer from the local directory
            _tokenizer = AutoTokenizer.from_pretrained(local_dir)
    except Exception as e:
        logger.error(f"An error occurred while initializing the tokenizer: {e}")


# utility functions
def load_config(config_file) -> Dict:
    """
    Load configuration from a local file or an S3 URI.

    :param config_file: Path to the local file or S3 URI (s3://bucket/key)
    :return: Dictionary with the loaded configuration
    """

    # Check if config_file is an S3 URI
    if config_file.startswith("s3://"):
        try:
            # Parse S3 URI
            s3_client = boto3.client('s3')
            bucket, key = config_file.replace("s3://", "").split("/", 1)

            # Get object from S3 and load YAML
            response = s3_client.get_object(Bucket=bucket, Key=key)
            return yaml.safe_load(response["Body"])
        except NoCredentialsError:
            print("AWS credentials not found.")
            raise
        except Exception as e:
            print(f"Error loading config from S3: {e}")
            raise
    # Check if config_file is an HTTPS URL
    elif config_file.startswith("https://"):
        try:
            response = requests.get(config_file)
            response.raise_for_status()  # Raises a HTTPError if the response was an error
            return yaml.safe_load(response.text)
        except requests.exceptions.RequestException as e:
            print(f"Error loading config from HTTPS URL: {e}")
            raise
    else:
        # Assume local file system if not S3 or HTTPS
        try:
            with open(config_file, 'r') as file:
                return yaml.safe_load(file)
        except Exception as e:
            print(f"Error loading config from local file system: {e}")
            raise
   
# The files in LongBench contain nonstandard or irregular Unicode.
# For compatibility and safety we normalize them.
def _normalize(text, form='NFC'):
    return unicodedata.normalize(form, text)

def _download_from_s3(bucket_name, prefix, local_dir):
    """Downloads files from an S3 bucket and a specified prefix to a local directory."""
    s3_client = boto3.client('s3')

    # Ensure the local directory exists
    if not os.path.exists(local_dir):
        os.makedirs(local_dir)

    # List and download files
    try:
        response = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
        if 'Contents' in response:
            for obj in response['Contents']:
                file_key = obj['Key']
                if file_key.endswith('/'):  
                    continue
                local_file_path = os.path.join(local_dir, os.path.basename(file_key))
                s3_client.download_file(bucket_name, file_key, local_file_path)
                logger.info(f"Downloaded: {local_file_path}")
        else:
            logger.warning(f"No files found in S3 Bucket: '{bucket_name}' with Prefix: '{prefix}'")
    except Exception as e:
        logger.error(f"An error occurred while downloading from S3: {e}")


def count_tokens(text: str) -> int:
    global _tokenizer
    _initialize_tokenizer()
    return len(_tokenizer.encode(text))

def process_item(item, prompt_fmt: str) -> Dict:
    question = _normalize(item.input)
    context = _normalize(item.context)
    prompt = prompt_fmt.format(question=question, context=context)
    prompt_len = count_tokens(prompt)
    ## generalize this further...
    ## bring your own script (if different) - bring your count token and your script
    return {
        "question": question,
        "context": context,
        "prompt": prompt,
        "prompt_len": prompt_len,
        "question_len": len(_tokenizer.encode(question)),
        "context_len": len(_tokenizer.encode(context)),
    }

def nt_to_posix(p: str) -> str:
    return p.replace("\\", "/")

# Function to write data to S3
def write_to_s3(json_data, bucket_name, dir1, dir2, file_name):

    # Initialize S3 client
    s3_client = boto3.client('s3')

    # Construct the S3 file path
    s3_file_path = posixpath.join(nt_to_posix(dir1), nt_to_posix(dir2), file_name)
    logger.info(f"write_to_s3, s3_file_path={s3_file_path}")
    try:
        # Write the JSON data to the S3 bucket
        s3_client.put_object(Bucket=bucket_name, Key=s3_file_path, Body=json_data)
        return (f"s3://{bucket_name}/{s3_file_path}")
    except NoCredentialsError:
        logger.error("write_to_s3, Error: AWS credentials not found.")
    except Exception as e:
        logger.error(f"write_to_s3, An error occurred: {e}")

        
## function to read from s3
def read_from_s3(bucket_name, s3_file_path):

    # Initialize S3 client
    s3_client = boto3.client('s3')
    s3_file_path = nt_to_posix(s3_file_path)

    try:
        # Fetch the object from S3
        logger.error(f"read_from_s3, reading file from bucket={bucket_name}, key={s3_file_path}")
        response = s3_client.get_object(Bucket=bucket_name, Key=s3_file_path)
        
        return response['Body'].read().decode('utf-8')
    except NoCredentialsError:
        logger.error("read_from_s3, Error: AWS credentials not found.")
        return None
    except Exception as e:
        logger.error(f"read_from_s3, An error occurred: {e}")
        return None

## gets a single s3 file
def get_s3_object(bucket: str, key: str) -> str:
 
    key = nt_to_posix(key)
    logger.info(f"get_s3_object, bucket_name={bucket}, key={key}")

    # Create an S3 client
    s3_client = boto3.client('s3')

    # Retrieve the object from S3
    response = s3_client.get_object(Bucket=bucket, Key=key)

    # Read the content of the file
    content = response['Body'].read().decode('utf-8')

    return content

# Function to list files in S3 bucket with a specific prefix
def list_s3_files(bucket, prefix, suffix='.json'):
    s3_client = boto3.client('s3')
    response = s3_client.list_objects_v2(Bucket=bucket, Prefix=nt_to_posix(prefix))
    return [item['Key'] for item in response.get('Contents', []) if item['Key'].endswith(suffix)]



# Import necessary libraries
import os
import time
import boto3
import logging
import tarfile
import sagemaker
from typing import Dict
from pathlib import Path
from urllib.parse import urlparse
from sagemaker.utils import name_from_base
from huggingface_hub import snapshot_download


## set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# globals
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf_token.txt")

## Initialize your S3 client for your model uploading
s3_client = boto3.client('s3')

## ------------- Use your specific execution role --------------------------------------------

# role=sagemaker.get_execution_role()  # execution role for the endpoint
sess=sagemaker.session.Session()  # sagemaker session for interacting with different AWS APIs
bucket=sess.default_bucket()  # bucket to house artifacts
model_bucket=sess.default_bucket()  # bucket to house artifacts

## ------------- Define the location of your s3 prefix for model artifacts ----------------------------------------

region=sess._region_name

## Define your account/session id
account_id=sess.account_id()

## ------------- Initialize the sagemaker and sagemaker runtime clients -------------------------------------------

sm_client = boto3.client("sagemaker")
smr_client = boto3.client("sagemaker-runtime")

## ---------------------------------------------------------------------------------------------------------

## Download the model snapshot
def download_model(experiment_config, model_name, local_model_path, allow_patterns):
    
    local_model_path = Path(local_model_path)
    print(f"Local model path: {local_model_path}")
    local_model_path.mkdir(exist_ok=True)
    print(f"Created the local directory: {local_model_path}")

    model_download_path = snapshot_download(
        repo_id=model_name,
        cache_dir=local_model_path,
        allow_patterns=allow_patterns,
        use_auth_token= Path(HF_TOKEN_FNAME).read_text().strip()
    )
    model_artifact = experiment_config['s3_path']
    print(f"Uncompressed model downloaded into ... -> {model_artifact}")
    return model_artifact

## Create the model artifact with the updated serving properties within the directory
def create_and_upload_model_artifact(serving_properties_path, bucket, key):
    # Create a tar.gz file containing only the serving.properties file
    tar_file_path = os.path.join(Path(serving_properties_path).parent, 'model.tar.gz')
    with tarfile.open(tar_file_path, "w:gz") as tar:
        # Add the serving.properties file
        tar.add(serving_properties_path, arcname='serving.properties')

    # Upload the tar.gz file to S3
    model_artifact_path = f"{key}/model.tar.gz"
    s3_client.upload_file(tar_file_path, bucket, model_artifact_path)
    model_tar_gz_path: str = f"s3://{bucket}/{model_artifact_path}"
    logger.info(f"uploaded model.tar.gz to {model_tar_gz_path}")
    return model_tar_gz_path

# Function to create the SageMaker model
def create_model(experiment_config, inference_image_uri, s3_model_artifact, role_arn):
    model_name = name_from_base(experiment_config['model_name'])
    create_model_response = sm_client.create_model(
        ModelName=model_name,
        ExecutionRoleArn=role_arn,
        PrimaryContainer={
            "Image": inference_image_uri,
            "ModelDataUrl": s3_model_artifact,
            "Environment": experiment_config['env'],
        },
    )
    return model_name, create_model_response["ModelArn"]

## Function to create and deploy the endpoint
def deploy_endpoint(experiment_config, model_name):
    endpoint_config_name = f"{model_name}-config"
    endpoint_name = f"{model_name}-endpoint"

    _ = sm_client.create_endpoint_config(
        EndpointConfigName=endpoint_config_name,
        ProductionVariants=[
            {
                "VariantName": "variant1",
                "ModelName": model_name,
                "InstanceType": experiment_config["instance_type"],
                "InitialInstanceCount": 1,
                "ModelDataDownloadTimeoutInSeconds": 3600,
                "ContainerStartupHealthCheckTimeoutInSeconds": 3600,
            },
        ],
    )
    
    create_endpoint_response = sm_client.create_endpoint(
        EndpointName=endpoint_name, EndpointConfigName=endpoint_config_name
    )
    return endpoint_name, create_endpoint_response['EndpointArn']

## Function to check the status of the endpoint
def check_endpoint_status(endpoint_name):
    resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
    status = resp["EndpointStatus"]
    while status == "Creating":
        time.sleep(60)
        resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
        status = resp["EndpointStatus"]
    return status

# Function to deploy the model and create the endpoint
def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    
    model_artifact = experiment_config['s3_path']
    print(f"Uncompressed model downloaded into ... -> {model_artifact}")
    
    logger.info("preparing model artifact...")

    # Set path to serving.properties (update this to the correct path)
    serving_properties_path: str = "p4d_serving.properties"
    dir_path = os.path.dirname(os.path.realpath(__file__))

    properties: str = Path(os.path.join(dir_path, serving_properties_path)).read_text()
    properties = properties.format(s3_path=experiment_config['s3_path'])
    serving_properties_path: str = os.path.join(dir_path, "serving.properties")
    Path(serving_properties_path).write_text(properties)

    # o = urlparse(experiment_config['s3_path'], allow_fragments=False)
    # bucket = o.netloc
    # logger.info(f"bucket name is -> {bucket}")
    # key = o.path
    # logger.info(f"key name is -> {key}")
    bucket = experiment_config['bucket_name']
    logger.info(f"bucket name is -> {bucket}")
    key = experiment_config['key_name']
    logger.info(f"key name is -> {key}")
    logger.info(f"uploading model to S3...bucket={bucket}, key={key}")
    model_artifact = create_and_upload_model_artifact(serving_properties_path, bucket, key)

    logger.info(f"Model uploaded to: {model_artifact}")

    inference_image_uri = experiment_config['image_uri']
    logger.info(f"Inference image URI: {inference_image_uri}")

    model_name, model_arn = create_model(experiment_config, inference_image_uri, model_artifact, role_arn)
    logger.info(f"Created Model: {model_arn}")

    endpoint_name, _ = deploy_endpoint(experiment_config, model_name)
    logger.info(f"Deploying Endpoint: {role_arn}")

    status = check_endpoint_status(endpoint_name)
    logger.info(f"Endpoint status: {status}")

    if status == 'InService':
        logger.info(f"endpoint is in service")
    else:
        logger.info("endpoint is not in service.")

    return dict(endpoint_name=endpoint_name, experiment_name=experiment_config['name'])